/*
*	© Copyright 2014, Group 4, ECE 492
*
*	This code was created through the collaboration
*	of Alix Krahn, Denis Lachance and Adam Thomson. 
*
*	This library takes in centroid objects and
*	tracks them as fingers. This way we can keep
*	track of which new centroid corresponds to
*	which finger. Tracks up to 10 fingers
*
*	type make to build, ./dsp to run
*	This was developed on Ubuntu 13.10
*/
#include "finger.h"

//finger objects
static struct Centroid fingers[FINGER_SIZE];
//count for how many frames were
//skipped for each finger. A threshold
//of 3 would allow 3 frames of blank
//values before discarding the finger
int finger_skip[FINGER_SIZE];

//* Function: Clear Fingers
//* Description: Clears all tracked fingers
//* Input: nothing
//* Returns: nothing
void ft_clear_fingers(){
	for (int i = 0; i < FINGER_SIZE; ++i)
	{
		fingers[i].x = 0;
		fingers[i].y = 0;
		fingers[i].size = 0;
	}
}

//* Function: Clear Fingers
//* Description: Clears all tracked fingers
//* Input: a Centroid array, length of said array
//* Returns: nothing
void ft_input_centroids(struct Centroid *cents, int len){
	int max_dist = DIST_THRESH*DIST_THRESH;
	int dist = max_dist;
	int min_i = -1;
	int tmp_dist = 0;
	int tmp_x, tmp_y;
	//bool array if centroid taken
	char taken[len];
	for (int i = 0; i < len; ++i)
	{
		taken[i] = 0;
	}
	//for all fingers that exist
	for (int i = 0; i < FINGER_SIZE; ++i)
	{
		//if exists
		if (fingers[i].size > 0)
		{
			//printf("exists!\n");
			//find closest centroid
			for (int j = 0; j < len; ++j)
			{
				//calulate squared distance (faster than regular distance)
				tmp_x = fingers[i].x - cents[j].x;
				tmp_y = fingers[i].y - cents[j].y;
				tmp_dist = tmp_x*tmp_x +tmp_y*tmp_y;
				//found minimum
				if (!taken[j] && tmp_dist < dist)
				{
					dist = tmp_dist;
					min_i = j;
				}
			}
			//assign to finger
			if (min_i >= 0)
			{
				//assign centroid to finger
				fingers[i] = cents[min_i];
				finger_skip[i] = 0;
				//centroid is taken now
				taken[min_i] = 1;
			}else{
				//increment the skipped frame counter
				finger_skip[i]++;
				//delete finger if we have passed the skipped frame threshold
				if (finger_skip[i] > SKIP_THRESH)
				{
					fingers[i].x = 0;
					fingers[i].y = 0;
					fingers[i].size = 0;
				}
			}
			//reset min index
			min_i = -1;
			//reset distance
			dist = max_dist;
		}
	}
	//for all leftover fingers
	for (int i = 0; i < FINGER_SIZE; ++i)
	{
		//if not exists
		if (fingers[i].size == 0)
		{
			//find an unused centroid
			for (int j = 0; j < len; ++j)
			{
				if (!taken[j] && cents[j].size > 0)
				{
					//assign unused centroid to blank finger
					fingers[i] = cents[j];
					taken[j] = 1;
					break;
				}
			}
		}
	}
}

//* Function: Get Finger
//* Description: Returns a finger at a specified
//index.
//* Input: index of finger
//* Returns: finger at index
struct Centroid ft_get_finger(int index){
	return fingers[index];
}